from typing import List

import torch

from .split import Layer, split_code
from .merge import merge_codebook


def match_cluster_centers(cluster_centers_ref: torch.Tensor, cluster_centers: torch.Tensor) -> torch.Tensor:
    '''
    Match the cluster centers to the reference.
    :param cluster_centers_ref: [K, C] float tensor, the reference cluster centers
    :param cluster_centers: [K, C] float tensor, the cluster centers to be matched
    :return: [K,] int tensor, the matched index
    '''
    assert cluster_centers_ref.shape == cluster_centers.shape
    return torch.argmin(torch.cdist(cluster_centers, cluster_centers_ref), dim=1)


def encode_known_layers(quantized_data: torch.Tensor, cluster_centers: torch.Tensor, layers: List[Layer]) -> List[Layer]:
    context = merge_codebook(layers)
    transcode = match_cluster_centers(context.cluster_centers, cluster_centers)
    codes = context.codebook[transcode[quantized_data]]
    code_layers = split_code(codes, n_bits=[layer.n_bit for layer in layers])
    return [layer._replace(codes=codes) for layer, codes in zip(layers, code_layers)]



from typing import Callable, List
import torch

from scalablevq.utils import tensor_bitcount


def n_bits_proposal_by_layer_scores(n_bit_baselayer: int, n_layers: int, layer_scores: torch.Tensor) -> List[int]:
    n_bits = [n_bit_baselayer]
    layer_scores_rest = layer_scores[n_bit_baselayer:]
    layer_score_proposal = layer_scores_rest.sum() / (n_layers - 1)
    for _ in range(n_layers):
        n_bit = 1
        while n_bit < layer_scores_rest.shape[0] and layer_scores_rest[:n_bit].sum() < layer_score_proposal:
            n_bit += 1
        n_bits.append(n_bit)
        layer_scores_rest = layer_scores_rest[n_bit:]
        if layer_scores_rest.shape[0] <= 0:
            break
    return n_bits


def n_bits_proposal_balanced_clusters(n_layers: int) -> Callable[[int, torch.Tensor, torch.Tensor], List[int]]:
    def n_bits_proposal(n_bit_baselayer: int, assigned_bits: torch.Tensor, _) -> List[int]:
        total_n_clusters = tensor_bitcount(assigned_bits).bincount()[1:]
        return n_bits_proposal_by_layer_scores(n_bit_baselayer, n_layers, total_n_clusters)
    # TODO: n_bits_proposal_by_layer_scores is not 100% accurate, as the score of a merged layer is not the sum of the scores of the two layers
    return n_bits_proposal


def n_bits_proposal_balanced_values(n_layers: int) -> Callable[[int, torch.Tensor, torch.Tensor], List[int]]:
    def n_bits_proposal(n_bit_baselayer: int, assigned_bits: torch.Tensor, values: torch.Tensor) -> List[int]:
        value_n_bits = tensor_bitcount(assigned_bits[values]) - 1
        n_values = value_n_bits.bincount()[1:]
        total_n_values = torch.cumsum(n_values.flip(dims=[0]), dim=0).flip(dims=[0])
        return n_bits_proposal_by_layer_scores(n_bit_baselayer, n_layers, total_n_values)
    return n_bits_proposal



from typing import List, Tuple
import open3d as o3d
import matplotlib
import numpy as np
import torch


def build_geometries(cluster_centers: torch.Tensor, layerized_cluster_centers: torch.Tensor, cluster_tree: List[Tuple[int, int]]):
    pcd_cluster_centers = o3d.geometry.PointCloud()
    pcd_cluster_centers.points = o3d.utility.Vector3dVector(cluster_centers.cpu())
    pcd_cluster_centers.colors = o3d.utility.Vector3dVector(np.array([[0, 0, 0]]*cluster_centers.shape[0]))

    leaf_n = cluster_centers.shape[0]
    lines = [(i+leaf_n, t[0]) for i, t in enumerate(cluster_tree)] + [(i+leaf_n, t[1]) for i, t in enumerate(cluster_tree)]
    line_set_layerized_cluster_centers = o3d.geometry.LineSet(
        points=o3d.utility.Vector3dVector(layerized_cluster_centers.cpu()),
        lines=o3d.utility.Vector2iVector(lines),
    )

    mat = o3d.visualization.rendering.MaterialRecord()
    mat.point_size = 5.0
    geometries = [
        {'name': 'cluster centers', 'geometry': pcd_cluster_centers, 'material': mat},
        {'name': 'layers', 'geometry': line_set_layerized_cluster_centers},
    ]
    return geometries


def build_clusters(data: torch.Tensor, quantized_data: torch.Tensor):
    geometries = []
    ids = quantized_data.unique()
    cmap = matplotlib.cm.get_cmap('rainbow')(np.linspace(0, 1, ids.shape[0]))
    for i in range(ids.shape[0]):
        cluster = data[quantized_data == ids[i], ...]
        pcd_cluster = o3d.geometry.PointCloud()
        pcd_cluster.points = o3d.utility.Vector3dVector(cluster.cpu())
        mat = o3d.visualization.rendering.MaterialRecord()
        mat.shader = 'defaultLitTransparency'
        mat.base_color = cmap[i, :3].tolist() + [0.8]
        mat.point_size = 1.0
        geometries.append({'name': f'cluster {i}', 'geometry': pcd_cluster, 'material': mat})
    return geometries


def visualize_layers(cluster_centers: torch.Tensor, layerized_cluster_centers: torch.Tensor, cluster_tree: List[Tuple[int, int]], data: torch.Tensor, quantized_data: torch.Tensor):
    o3d.visualization.draw(build_geometries(cluster_centers, layerized_cluster_centers, cluster_tree) + build_clusters(data, quantized_data))



import math
from typing import Callable, List

import torch

from .build import build_layers, vdistance
from .assign import assign_bits
from .utils import bitcount
from .split import Layer, split_layers


def format_n_bits(n_bits_proposal: List[int], max_n_bits) -> str:
    n_bits = []
    while n_bits_proposal and max_n_bits > 0:
        n_bit = min(n_bits_proposal[0], max_n_bits)
        n_bits_proposal = n_bits_proposal[1:]
        n_bits.append(n_bit)
        max_n_bits -= n_bit
    if max_n_bits > 0:
        n_bits.append(max_n_bits)
    return n_bits


def encode_layers(
        data: torch.Tensor, quantized_data: torch.Tensor, cluster_centers: torch.Tensor,
        n_bit_baselayer: int = 4, n_bits_proposal: int | List[int] | Callable[[int, torch.Tensor, torch.Tensor], List[int]] = [4, 4, 4],
        n_bit_limit: int = 63, dist_func=vdistance, visualize=False) -> List[Layer]:
    '''
    Encode the layers.
    :param: data: [N, C] float tensor, the origional data
    :param: quantized_data: [N] long tensor, the quantized data
    :param: cluster_centers: [K, C] float tensor, the cluster centers
    :param: n_bits_proposal: list of int, bit width proposal for each layer
    :return: [layer0, layer1, ...] splitted layers
    '''
    assert quantized_data.unique().shape[0] == cluster_centers.shape[0], "No empty cluster is allowed"
    assert 2**n_bit_baselayer < cluster_centers.shape[0], "The number of clusters should be more than 2**n_bit_baselayer"

    layerized_cluster_centers, cluster_tree = build_layers(
        cluster_centers, data, quantized_data,
        final_clusters=2**n_bit_baselayer,
        depth_limit=n_bit_limit - n_bit_baselayer,
        dist_func=dist_func)
    if visualize:
        from .visualize import visualize_layers
        visualize_layers(cluster_centers, layerized_cluster_centers, cluster_tree, data=data, quantized_data=quantized_data)
    assigned_bits = assign_bits(cluster_tree, n_clusters=layerized_cluster_centers.shape[0])
    assigned_bits = torch.tensor(assigned_bits, device=quantized_data.device)
    assigned_bits_data = assigned_bits[quantized_data]

    max_n_bits = bitcount(assigned_bits.max().item()) - 1
    if isinstance(n_bits_proposal, int):
        n_bits = format_n_bits([n_bit_baselayer] + [n_bits_proposal] * math.ceil((max_n_bits - n_bit_baselayer) / n_bits_proposal), max_n_bits)
    elif isinstance(n_bits_proposal, list):
        n_bits = format_n_bits([n_bit_baselayer] + n_bits_proposal, max_n_bits)
    else:
        n_bits = format_n_bits(n_bits_proposal(n_bit_baselayer, assigned_bits, quantized_data), max_n_bits)

    leaf_idx = torch.arange(0, cluster_centers.shape[0], dtype=assigned_bits.dtype, device=assigned_bits_data.device)
    layers = split_layers(assigned_bits_data, assigned_bits, leaf_idx, layerized_cluster_centers, n_bits=n_bits)
    return layers



from typing import List, Tuple, NamedTuple

import torch

from .split import Layer


class MergingContext(NamedTuple):
    codebook: torch.Tensor  # [K,] int tensor
    cluster_centers: torch.Tensor  # [K, C] float tensor, with the same order as codebook
    n_leaf: int  # the number of leaf codes for this layer, codebook[:n_leaf] are leaf codes

    sorted_codebook: torch.Tensor
    sorted_codebook_idx: torch.Tensor


def construct_context(codebook: torch.Tensor, cluster_centers: torch.Tensor, n_leaf: int) -> MergingContext:
    sorted_codebook, sorted_idx = codebook.sort()
    return MergingContext(codebook, cluster_centers, n_leaf, sorted_codebook, sorted_idx)


def init_codebook(layer: Layer) -> MergingContext:
    return construct_context(layer.codebook, layer.cluster_centers, layer.n_leaf)


def append_codebook(layer: Layer, context: MergingContext) -> MergingContext:
    codebook = torch.cat([context.codebook[:context.n_leaf], layer.codebook])
    cluster_centers = torch.cat([context.cluster_centers[:context.n_leaf], layer.cluster_centers])
    n_leaf = context.n_leaf + layer.n_leaf
    return construct_context(codebook, cluster_centers, n_leaf)


def merge_init(layer: Layer) -> Tuple[torch.Tensor, MergingContext]:
    '''
    Init the merging by the lowest layer.
    :param layer: the lowest layer.
    :return: the codes and the context.
    '''
    codes = layer.codes + (1 << layer.n_bit)
    context = init_codebook(layer)
    return codes, context


def merge_next_layer(layer: Layer, codes: torch.Tensor, context: MergingContext) -> Tuple[torch.Tensor, MergingContext]:
    '''
    Merging the next layer.
    :param layer: the next layer.
    :param codes: the current codes.
    :param context: the current context.
    :return: the new codes and the new context.
    '''
    is_leaf = context.sorted_codebook_idx[torch.searchsorted(context.sorted_codebook, codes)] < context.n_leaf
    codes[~is_leaf] = (codes[~is_leaf] << layer.n_bit) + layer.codes
    context = append_codebook(layer, context)
    return codes, context


def merge_layers(layers: List[Layer]) -> Tuple[torch.Tensor, MergingContext]:
    '''
    Merge multiple layers.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: the codes and the context.
    '''
    codes, context = merge_init(layers[0])
    for layer in layers[1:]:
        codes, context = merge_next_layer(layer, codes, context)
    return codes, context


def merge_codebook(layers: List[Layer]) -> MergingContext:
    '''
    Merge only the codebook.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: the context.
    '''
    context = init_codebook(layers[0])
    for layer in layers[1:]:
        context = append_codebook(layer, context)
    return context



from typing import List, Tuple, Callable
import torch
import tqdm
import math


def split_into_clusters(n_clusters: int, data: torch.Tensor, quantized_data: torch.Tensor):
    """Split the data into clusters"""
    clusters = [None]*n_clusters
    for i in tqdm.tqdm(range(n_clusters), desc="Split data into clusters"):
        clusters[i] = data[quantized_data == i, ...]
    return clusters


def knn_cluster_centers(cluster_centers: torch.Tensor, n=2, batch=2**16):
    """Get k neighbor points for each point."""
    neighbors_idx = torch.zeros(cluster_centers.shape[0], n, dtype=torch.int64, device=cluster_centers.device)
    progress_bar = tqdm.tqdm(range(cluster_centers.shape[0]), desc="Init K-Nearest for cluster centers")
    dists = torch.zeros(cluster_centers.shape[0], n, dtype=cluster_centers.dtype, device=cluster_centers.device)
    for i in range(math.ceil(cluster_centers.shape[0]/batch)):
        dist = torch.norm(cluster_centers[i*batch:i*batch+batch, ...].unsqueeze(-2) - cluster_centers, p=2, dim=-1)
        knn = dist.topk(n + 1, largest=False)
        dists[i*batch:i*batch+batch, ...] = knn.values[:, 1:]
        neighbors_idx[i*batch:i*batch+batch, ...] = knn.indices[:, 1:]
        progress_bar.update(min(i*batch+batch, cluster_centers.shape[0])-i*batch)
    return neighbors_idx, dists


def merged_cluster(a: torch.Tensor, b: torch.Tensor):
    ab = torch.cat([a, b], dim=0)
    center = ab.mean(dim=0, keepdim=True)
    return ab, center


def vdistance(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """Virtual distance for quant cluster_tree. Just the avg distance from center to points after merge."""
    ab, center = merged_cluster(a, b)
    dist = torch.norm(ab-center, dim=1, p=2).mean()
    return dist


def distance_init(clusters: List[torch.Tensor], cluster_centers: torch.Tensor, neighbors_idx: torch.Tensor,
                  dist_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = vdistance):
    """Calculate the virtual distance between the point and their k neighbor points"""
    dists = torch.zeros(neighbors_idx.shape, dtype=cluster_centers.dtype, device=cluster_centers.device)
    for center_idx in tqdm.tqdm(range(cluster_centers.shape[0]), desc="Compute 'distance' among K-Nearest"):
        data_dequant = clusters[center_idx]
        for i, neighbor_idx in enumerate(neighbors_idx[center_idx, ...]):
            data_neighbor_dequant = clusters[neighbor_idx]
            dists[center_idx, i] = dist_func(data_dequant, data_neighbor_dequant)
    return dists


def merge(
    layerized_cluster_centers: torch.Tensor, cluster_tree: List[Tuple[int, int]],
    clusters: List[torch.Tensor], dists: torch.Tensor,
    tmp_cluster_centers: torch.Tensor, tmp_neighbors_idx: torch.Tensor,
    tmp_depth: torch.Tensor, depth_limit: int = 64,
    dist_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = vdistance
):
    """Merge the clusters and dists, and expand layerized_cluster_centers"""
    min_pos = (dists == dists.min()).nonzero()[0]
    min_idx, min_neighbors_idx = min_pos[0].item(), tmp_neighbors_idx[*min_pos].item()
    # concat to get new cluster and center
    new_cluster, new_center = merged_cluster(clusters[min_idx], clusters[min_neighbors_idx])
    new_depth = max(tmp_depth[min_idx], tmp_depth[min_neighbors_idx]) + 1
    # add new cluster
    new_center_idx = len(clusters)
    clusters.append(None)
    clusters[new_center_idx] = new_cluster if new_depth < depth_limit else None  # control the tree depth
    # delete old clusters
    clusters[min_idx] = clusters[min_neighbors_idx] = None
    # add new center
    tmp_cluster_centers = torch.cat([tmp_cluster_centers, new_center], dim=0)
    tmp_cluster_centers[-1, ...] = new_center if new_depth < depth_limit else torch.inf  # control the tree depth
    tmp_depth = torch.cat([tmp_depth, torch.tensor([new_depth], device=tmp_depth.device)], dim=0)
    layerized_cluster_centers = torch.cat([layerized_cluster_centers, new_center], dim=0)  # save result
    cluster_tree.append((min_idx, min_neighbors_idx))  # save result
    # delete old center
    tmp_cluster_centers[min_idx, ...] = torch.inf
    tmp_cluster_centers[min_neighbors_idx, ...] = torch.inf
    # which center should be updated
    should_update_idx = torch.logical_or(
        tmp_neighbors_idx == min_idx,
        tmp_neighbors_idx == min_neighbors_idx
    ).any(dim=1).nonzero()[..., 0]
    should_update_idx = should_update_idx[torch.logical_and(
        should_update_idx != min_idx, should_update_idx != min_neighbors_idx
    )]
    should_update_idx = torch.cat([
        should_update_idx,
        torch.tensor([new_center_idx], dtype=should_update_idx.dtype, device=should_update_idx.device)
    ], dim=0)
    # compute new neighbors idx
    new_kdist = torch.norm(tmp_cluster_centers[should_update_idx, ...].unsqueeze(-2) - tmp_cluster_centers, p=2, dim=-1)
    new_knn = new_kdist.topk(dists.shape[1] + 1, largest=False)
    # update neighbors idx
    tmp_neighbors_idx = torch.cat([tmp_neighbors_idx, torch.zeros_like(tmp_neighbors_idx[0:1, ...])], dim=0)
    tmp_neighbors_idx[should_update_idx, ...] = new_knn.indices[:, 1:]
    # update neighbors distance
    dists = torch.cat([dists, torch.zeros_like(dists[0:1, ...])], dim=0)
    for center_idx in tqdm.tqdm(should_update_idx, desc="Updating distance", position=1, leave=False):
        center_data = clusters[center_idx]
        for i, neighbor_idx in enumerate(tmp_neighbors_idx[center_idx, ...]):
            neighbor_data = clusters[neighbor_idx]
            if center_data is not None and neighbor_data is not None:
                dists[center_idx, i] = dist_func(center_data, neighbor_data)
            else:
                dists[center_idx, i] = torch.inf
    dists[min_idx, ...] = torch.inf
    dists[min_neighbors_idx, ...] = torch.inf
    return layerized_cluster_centers, cluster_tree, clusters, dists, tmp_cluster_centers, tmp_neighbors_idx, tmp_depth


def build_layers(cluster_centers: torch.Tensor, data: torch.Tensor, quantized_data: torch.Tensor, final_clusters: int, depth_limit=64,
                 dist_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = vdistance):
    clusters = split_into_clusters(cluster_centers.shape[0], data, quantized_data)
    neighbors_idx, _ = knn_cluster_centers(cluster_centers)
    dists = distance_init(clusters, cluster_centers, neighbors_idx, dist_func=dist_func).to(data.device)
    layerized_cluster_centers = cluster_centers.clone()
    cluster_tree = []
    tmp_cluster_centers = cluster_centers.clone()
    tmp_neighbors_idx = neighbors_idx.clone()
    tmp_depth = torch.ones(cluster_centers.shape[0], dtype=torch.int, device=cluster_centers.device)
    context = (layerized_cluster_centers, cluster_tree, clusters, dists, tmp_cluster_centers, tmp_neighbors_idx, tmp_depth)
    for _ in tqdm.tqdm(range(len(clusters)-final_clusters), desc="Merging clusters", position=0, leave=False):
        context = merge(*context, depth_limit=depth_limit, dist_func=dist_func)
        torch.cuda.empty_cache()
    (layerized_cluster_centers, cluster_tree, clusters, dists, tmp_cluster_centers, tmp_neighbors_idx, tmp_depth) = context
    return layerized_cluster_centers, cluster_tree



import torch
import numpy as np


def bitcount(a: int):
    max_n_bits = int(np.ceil(np.log2(a)))
    counts = 0
    for i in range(0, max_n_bits):
        counts += 1 if (a >> i) > 0 else 0
    return counts


def tensor_bitcount(a: torch.Tensor):
    max_n_bits = int(torch.ceil(torch.log2(a.max())))
    counts = torch.zeros_like(a)
    for i in range(0, max_n_bits):
        counts[(a >> i) > 0] += 1
    return counts



from .build import build_layers
from .assign import assign_bits
from .split import split_code, split_codebook, split_layers, Layer
from .encode import format_n_bits, encode_layers
from .decode import extract_layers, decode_layers, extract_layer, decode_layer
from .nbits import n_bits_proposal_balanced_clusters, n_bits_proposal_balanced_values
from .known import encode_known_layers



from typing import List, Tuple
import numpy as np
from .utils import bitcount


def assign_bits(tree: List[Tuple[int, int]], n_clusters: int):
    n_clusters_init = n_clusters - len(tree)
    assigned = [[] for _ in range(n_clusters)]
    top = [True]*n_clusters

    def append_higherbits_until_leaf(k, bits):
        assigned[k] = bits + assigned[k]
        if k < n_clusters_init:
            return
        append_higherbits_until_leaf(tree[k-n_clusters_init][0], bits)
        append_higherbits_until_leaf(tree[k-n_clusters_init][1], bits)

    for l, r in tree:
        append_higherbits_until_leaf(l, [0])
        append_higherbits_until_leaf(r, [1])
        top[l] = top[r] = False

    top_idx, = np.where(top)
    lod0_bitwidth = bitcount(len(top_idx))
    for b, j in enumerate(top_idx):
        append_higherbits_until_leaf(j, [int(bit) for bit in format(b, f'0{lod0_bitwidth}b')])
    assigned_bits = [eval('0b' + ''.join([str(1)] + [str(bit) for bit in bits])) for bits in assigned]  # 1 on the hightest bit
    return assigned_bits



from typing import List, NamedTuple, Tuple

import torch

from .utils import tensor_bitcount


def split_code(codes: torch.Tensor, n_bits: List[int] = [4, 4, 4]) -> List[torch.Tensor]:
    '''
    Split the encoded codes into multiple chunks.
    :param codes: [N,] int tensor, the codes to be split
    :param n_bits: [n1, n2, n3] int list, the number of code bits for each chunks
    :return: [chunk1, chunk2, ...] each chunk is a 1-dim int tensor
    '''
    chunks = []
    rest_codes = codes.clone()
    bitlength = tensor_bitcount(rest_codes) - 1
    for i in range(len(n_bits)):
        n_bit = n_bits[i]
        end_bit = sum(n_bits[:i + 1])
        n_clip_bit = bitlength - end_bit
        cliped_rest_codes = rest_codes >> n_clip_bit
        cliped_rest_codes[n_clip_bit < 0] = rest_codes[n_clip_bit < 0] << -n_clip_bit[n_clip_bit < 0]
        chunk = cliped_rest_codes & ((1 << n_bit) - 1)  # split out 1 layer
        done_mask = n_clip_bit <= 0
        rest_codes = rest_codes[~done_mask]  # collect rest codes
        bitlength = bitlength[~done_mask]
        chunks.append(chunk)
    return chunks


def format_leaf(codebook_leaf: torch.Tensor, n_bits: List[int]) -> torch.Tensor:
    bitlength = tensor_bitcount(codebook_leaf) - 1
    for i in range(len(n_bits)):
        end_bit = sum(n_bits[:i + 1])
        n_clip_bit = bitlength - end_bit
        move_idx = (n_clip_bit < 0) & (n_clip_bit > -n_bits[i])
        codebook_leaf[move_idx] = codebook_leaf[move_idx] << -n_clip_bit[move_idx]
    return codebook_leaf


def reverse_codebook(codebook: torch.Tensor) -> torch.Tensor:
    reversed_codebook = torch.zeros(codebook.max() + 1, dtype=codebook.dtype, device=codebook.device) + len(codebook)
    reversed_codebook[codebook] = torch.arange(len(codebook), dtype=codebook.dtype, device=codebook.device)
    return reversed_codebook


def split_codebook(codebook: torch.Tensor, leaf_idx: torch.Tensor, cluster_centers: torch.Tensor, n_bits: List[int] = [4, 4, 4]) -> List[Tuple[torch.Tensor, torch.Tensor]]:
    '''
    Split the codebook for multiple chunks.
    :param codebook: [N,] int tensor, all the codes
    :param leaf_idx: [M,] int tensor, identify the index of leaf codes in the codebook
    :param cluster_centers: [N, C] cluster centers with the same order as codebook
    :param n_bits: [n1, n2, n3] int list, the number of code bits for each chunks
    :return: [(codebook1, cluster_centers1, n_leaf1), (cb2, cc2, nl2), (cb3, cc3, nl3), ...] each chunk is a 1-dim int tensor
    '''
    codebook_leaf = format_leaf(codebook[leaf_idx], n_bits)
    codebook = codebook.clone()
    codebook[leaf_idx] = codebook_leaf
    sorted_codebook, sorted_idx = codebook.sort()
    is_leaf = torch.zeros(codebook.shape[0], dtype=torch.bool, device=codebook.device)
    is_leaf[leaf_idx] = True

    bitlength = tensor_bitcount(codebook_leaf) - 1
    chunks = []
    for i in range(len(n_bits)):
        end_bit = sum(n_bits[:i + 1])
        n_clip_bit = bitlength - end_bit
        chunk = codebook_leaf >> n_clip_bit
        codebook_layer = chunk.unique()

        is_leaf_layer = is_leaf[sorted_idx[torch.searchsorted(sorted_codebook, codebook_layer)]]
        codebook_layer = torch.cat([codebook_layer[is_leaf_layer], codebook_layer[~is_leaf_layer]])
        n_leaf = is_leaf_layer.sum().item()  # put non-leaf codes in the end, convenient for merging when decoding
        cluster_centers_layer = cluster_centers[sorted_idx[torch.searchsorted(sorted_codebook, codebook_layer)], ...]

        chunks.append((codebook_layer, cluster_centers_layer, n_leaf))

        done_mask = n_clip_bit <= 0
        codebook_leaf = codebook_leaf[~done_mask]
        bitlength = bitlength[~done_mask]
    assert codebook_leaf.shape[0] == 0
    return chunks


class Layer(NamedTuple):
    codes: torch.Tensor  # [N,] int tensor, code.max() < K
    codebook: torch.Tensor  # [K,] int tensor
    cluster_centers: torch.Tensor  # [K, C] float tensor, with the same order as codebook
    n_bit: int  # the number of code bits for this layer
    n_leaf: int  # the number of leaf codes for this layer, codebook[:n_leaf] are leaf codes


def split_layers(codes: torch.Tensor, codebook: torch.Tensor, leaf_idx: torch.Tensor, cluster_centers: torch.Tensor, n_bits: List[int] = [4, 4, 4]) -> List[Layer]:
    '''
    Split the encoded codes and codebook into multiple layers.
    :param codes: [N,] int tensor, the codes to be split
    :param codebook: [N,] int tensor, all the codes
    :param leaf_idx: [M,] int tensor, identify the index of leaf codes in the codebook
    :param cluster_centers: [N, C] cluster centers with the same order as codebook
    :param n_bits: [n1, n2, n3] int list, the number of code bits for each chunks
    :return: [layer0, layer1, ...] splitted layers
    '''
    code_layers = split_code(codes, n_bits)
    codebook_layers = split_codebook(codebook, leaf_idx, cluster_centers, n_bits)
    return [Layer(code, codebook, cluster_centers, n_bit, n_leaf) for code, (codebook, cluster_centers, n_leaf), n_bit in zip(code_layers, codebook_layers, n_bits)]



from typing import List, NamedTuple, Tuple

import torch

from .split import Layer
from .merge import MergingContext, merge_layers, merge_init, merge_next_layer


def extract_layers(layers: List[Layer]) -> Tuple[torch.Tensor, torch.Tensor]:
    '''
    Decode the layers.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: ([N, C], [K, C]) int tensor ids, float tensor cluster centers
    '''
    codes, context = merge_layers(layers)
    return context.sorted_codebook_idx[torch.searchsorted(context.sorted_codebook, codes)], context.cluster_centers


def decode_layers(layers: List[Layer]) -> torch.Tensor:
    '''
    Decode the layers.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: [N, C] float tensor
    '''
    ids, cluster_centers = extract_layers(layers)
    return cluster_centers[ids, ...]


class DecodingContext(NamedTuple):
    codes: torch.Tensor
    merging: MergingContext


def extract_layer(layer: Layer, context: DecodingContext = None) -> Tuple[torch.Tensor, torch.Tensor, DecodingContext]:
    '''
    Decode the layers.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: the decode result ([N, C] int tensor ids, [K, C] float tensor cluster centers) and the new context
    '''
    if context is None:
        codes, merging = merge_init(layer)
    else:
        codes, merging = merge_next_layer(layer, context.codes, context.merging)
    return merging.sorted_codebook_idx[torch.searchsorted(merging.sorted_codebook, codes)], merging.cluster_centers, DecodingContext(codes, merging)


def decode_layer(layer: Layer, context: DecodingContext = None) -> Tuple[torch.Tensor, DecodingContext]:
    '''
    Decode the layers.
    :param: layers: [layer0, layer1, ...] splitted layers
    :return: the decode result and the new context
    '''
    ids, cluster_centers, context = extract_layer(layer, context)
    return cluster_centers[ids, ...], context
